# %%

import numpy as np
import matplotlib.pyplot as plt
import matplotlib
from .core.utility import Parameters
from .core.tessellation import Tessellation
from types import ModuleType


class Cpab:
    """Continous piecewise-affine based transformations. Package main class.

    Args:
        tess_size (int): tesselation size, with the number of vertices.
        backend (str, optional): computational backend to use. Choose between "*numpy*", or "*pytorch*". Defaults to "*numpy*".
        device (str, optional): device to use. Choose between "*cpu*" or "*gpu*". For the numpy backend only the "*cpu*" option is valid. Defaults to "*cpu*".
        zero_boundary (bool, optional): determines if the velocity at the boundary is zero 
        basis (str, optional): constrain basis to use. Choose between "*rref*", "*svd*", or "*sparse*". Defaults to "*rref*".
    """

    params: Parameters
    tess: Tessellation
    backend_name: str
    backend: ModuleType

    def __init__(
        self,
        tess_size: int,
        backend: str = "numpy",
        device: str = "cpu",
        zero_boundary: bool = True,
        basis: str = "rref",
    ):
        # Check input
        self._check_input(tess_size, backend, device, zero_boundary, basis)

        # Parameters
        self.params = Parameters()
        self.params.nc = tess_size
        self.params.zero_boundary = zero_boundary
        self.params.xmin = 0
        self.params.xmax = 1
        self.params.nSteps1 = 10
        self.params.nSteps2 = 5
        self.params.precomputed = False
        self.params.use_slow = False
        self.params.basis = basis

        # Initialize tesselation
        self.tess = Tessellation(
            self.params.nc, self.params.xmin, self.params.xmax, self.params.zero_boundary, basis=self.params.basis,
        )

        # Extract parameters from tesselation
        self.params.B = self.tess.B
        self.params.D, self.params.d = self.tess.D, self.tess.d

        self.backend_name = backend

        # Load backend and set device
        self.device = device.lower()
        self.backend_name = backend
        if self.backend_name == "numpy":
            from .backend.numpy import functions as backend
        elif self.backend_name == "pytorch":
            from .backend.pytorch import functions as backend

            self.params.B = backend.to(self.params.B, device=self.device)
            self.params.B = self.params.B.contiguous()
        self.backend = backend

        # Assert that we have a recent version of the backend
        self.backend.assert_version()

    def uniform_meshgrid(self, n_points: int):
        """Generates a uniformly separated, one-dimensional meshgrid 

        Args:
            n_points (int): number of points

        Returns:
            numpy.ndarray or torch.Tensor: vector of points
        """
        return self.backend.uniform_meshgrid(self.params.xmin, self.params.xmax, n_points, self.device)

    def covariance_cpa(self, length_scale: float = 0.1, output_variance: float = 1):
        """Generates the covariance matrix for a continuous piecewise-affine space. 
        This function uses the squared exponential kernel with parameters 
        :math:`\lambda_s`: *length_scale* and :math:`\lambda_v`: *output_variance*
        
        Args:
            length_scale (float,>0): how fast the covariance declines between the cells 
            output_variance (float,>0): overall variance from the mean
        
        Returns:
            numpy.ndarray or torch.Tensor: [d, d] matrix. 
        """

        centers = self.backend.to(self.tess.cell_centers(), device=self.device)
        dist = self.backend.pdist(centers)

        cov_init = self.backend.ones((self.params.D, self.params.D))
        cov_init = self.backend.to(cov_init, device=self.device) * 100 * self.backend.max(dist)
        cov_init[::2, ::2] = dist
        cov_init[1::2, 1::2] = dist

        # Squared exponential kernel + Smoothness priors on CPA
        cov_pa = output_variance ** 2 * self.backend.exp(-(cov_init / (2 * length_scale ** 2)))

        B = self.params.B
        cov_cpa = self.backend.matmul(self.backend.matmul(B.T, cov_pa), B)

        return self.backend.to(cov_cpa, device=self.device)

    def sample_transformation(self, n_sample: int, mean=None, cov=None):
        """Generates samples of a random transformation vector. By default the method samples from a standard Gaussian distribution.
        
        Args:
            n_sample (int): number of transformations to sample
            mean (numpy.ndarray or torch.Tensor): [d,] vector, mean of multivariate gaussian
            cov (numpy.ndarray or torch.Tensor): [d,d] matrix, covariance of multivariate gaussian

        Returns:
            numpy.ndarray or torch.Tensor: [n_sample, d] matrix, where each row is a independent sample from a multivariate gaussian
        """
        if mean is not None:
            self._check_type(mean)
            self._check_device(mean)
        if cov is not None:
            self._check_type(cov)
            self._check_device(cov)

        samples = self.backend.sample_transformation(self.params.d, n_sample, mean, cov, self.device)

        return self.backend.to(samples, device=self.device)

    def sample_transformation_with_prior(
        self, n_sample: int, mean=None, length_scale: float = 0.1, output_variance: float = 1
    ):
        """Generates samples of smooth transitions. Covariance matrix for the continuous piecewise-affine space. 
        This function uses the squared exponential kernel with parameters 
        :math:`\lambda_s`: *length_scale* and :math:`\lambda_v`: *output_variance*
        
        Args:
            n_sample (int): number of transformations to sample
            mean (numpy.ndarray or torch.Tensor): [d,] vector, mean of multivariate gaussian
            length_scale (float,>0): how fast the covariance declines between the cells 
            output_variance (float,>0): overall variance from the mean

        Returns:
            numpy.ndarray or torch.Tensor: [d, d] matrix. 
        """

        if mean is not None:
            self._check_type(mean)
            self._check_device(mean)

        centers = self.backend.to(self.tess.cell_centers(), device=self.device)
        dist = self.backend.pdist(centers)

        cov_init = self.backend.ones((self.params.D, self.params.D))
        cov_init = self.backend.to(cov_init, device=self.device) * 100 * self.backend.max(dist)
        cov_init[::2, ::2] = dist
        cov_init[1::2, 1::2] = dist

        # Squared exponential kernel + Smoothness priors on CPA
        cov_pa = output_variance ** 2 * self.backend.exp(-(cov_init / (2 * length_scale ** 2)))

        B = self.params.B
        cov_cpa = self.backend.matmul(self.backend.matmul(B.T, cov_pa), B)

        samples = self.backend.sample_transformation(self.params.d, n_sample, mean, cov_cpa, self.device)
        return self.backend.to(samples, device=self.device)

    def identity(self, n_sample: int = 1, epsilon: float = 0):
        """Generates the parameters for the identity transformation (vector of zeros)

        Args:
            n_sample (int): number of transformations to sample
            epsilon (float, >0): number to add to the identity transformation for stability 
        
        Returns:
            numpy.ndarray or torch.Tensor: [n_sample, d] matrix, where each row is a sample
        """
        return self.backend.identity(self.params.d, n_sample, epsilon, self.device)

    def transform_grid(self, grid, theta, method=None, time=1.0):
        """Integrates a grid using the parametrization in theta.

        Args:
            grid (numpy.ndarray or torch.Tensor): [n_points] vector or [n_batch, n_points] tensor i.e.
                either a single grid for all theta values, or a grid for each theta value
            theta (numpy.ndarray or torch.Tensor): [n_batch, d] matrix
            method (str): integration method, one of "*closed_form*", "*numeric*".
            time (float): integration time. Defaults to 1.0
        
        Returns:
            numpy.ndarray or torch.Tensor: [n_batch, n_points] tensor, with the transformed grid.
            The slice transformed_grid[i] corresponds to the grid being transformed by theta[i]
        """
        self._check_type(grid)
        self._check_device(grid)
        self._check_type(theta)
        self._check_device(theta)
        transformed_grid = self.backend.transformer(grid, theta, self.params, method, time)
        return transformed_grid

    def transform_grid_ss(self, grid, theta, method=None, time=1.0, N=0):
        """Integrates a grid using the scaling-squaring method and the parametrization in theta.

        Args:
            grid (numpy.ndarray or torch.Tensor): [n_points] vector or [n_batch, n_points] tensor i.e.
                either a single grid for all theta values, or a grid for each theta value
            theta (numpy.ndarray or torch.Tensor): [n_batch, d] matrix
            method (str): integration method, one of "*closed_form*", "*numeric*".
            time (float): integration time. Defaults to 1.0
            N (int): number of scaling iterations. Defaults to 0
        
        Returns:
            numpy.ndarray or torch.Tensor: [n_batch, n_points] tensor, with the transformed grid.
            The slice transformed_grid[i] corresponds to the grid being transformed by theta[i]
        """

        self._check_type(grid)
        self._check_device(grid)
        self._check_type(theta)
        self._check_device(theta)
        # time = time / 2**N
        transformed_grid = self.backend.transformer(grid, theta, self.params, method, time)
        for i in range(N):
            transformed_grid = self.backend.interpolate_grid(transformed_grid, self.params)
        return transformed_grid

    def gradient_grid(self, grid, theta, method=None, time=1.0):
        """ Integrates and returns the gradient of the transformation w.r.t. the parametrization in theta.

        Args:
            grid (numpy.ndarray or torch.Tensor): [n_points] vector or [n_batch, n_points] tensor i.e.
                either a single grid for all theta values, or a grid for each theta value
            theta (numpy.ndarray or torch.Tensor): [n_batch, d] matrix
        
        Returns:
            tuple: tuple containing
                - **transformed_grid** (numpy.ndarray or torch.Tensor): [n_batch, n_points] tensor, with the transformed grid. The slice transformed_grid[i] corresponds to the grid being transformed by theta[i]
                
                - **gradient_grid** (numpy.ndarray or torch.Tensor): [n_batch, n_points, d] tensor, with the gradient grid. The slice gradient_grid[i, j] corresponds to the gradient of grid being transformed by theta[i] and with respect to the parameter theta[j]
            
        """
        self._check_type(grid)
        self._check_device(grid)
        self._check_type(theta)
        self._check_device(theta)
        gradient_grid = self.backend.gradient(grid, theta, self.params, method, time)
        return gradient_grid

    def gradient_space(self, grid, theta, method=None, time=1.0):
        """ Integrates and returns the gradient of the transformation w.r.t. the parametrization in theta.

        Args:
            grid (numpy.ndarray or torch.Tensor): [n_points] vector or [n_batch, n_points] tensor i.e.
                either a single grid for all theta values, or a grid for each theta value
            theta (numpy.ndarray or torch.Tensor): [n_batch, d] matrix
        
        Returns:
            tuple: tuple containing
                - **transformed_grid** (numpy.ndarray or torch.Tensor): [n_batch, n_points] tensor, with the transformed grid. The slice transformed_grid[i] corresponds to the grid being transformed by theta[i]
                
                - **gradient_grid** (numpy.ndarray or torch.Tensor): [n_batch, n_points, d] tensor, with the gradient grid. The slice gradient_grid[i, j] corresponds to the gradient of grid being transformed by theta[i] and with respect to the parameter theta[j]
            
        """
        self._check_type(grid)
        self._check_device(grid)
        self._check_type(theta)
        self._check_device(theta)
        gradient_space = self.backend.gradient_space(grid, theta, self.params, method, time)
        return gradient_space

    def interpolate(self, data, grid, outsize: int):
        """ Linear interpolation method

        Args:
            data (numpy.ndarray or torch.Tensor): [n_batch, n_points, n_channels] tensor with input data.
            grid (numpy.ndarray or torch.Tensor): [n_batch, n_points] tensor with grid points that are used to interpolate the data
            outsize: number of points in the output
        
        Returns:
            interpolated(numpy.ndarray or torch.Tensor): [n_batch, outsize, n_channels] tensor with the interpolated data
        """
        self._check_type(data)
        self._check_device(data)
        self._check_type(grid)
        self._check_device(grid)
        return self.backend.interpolate(data, grid, outsize)

    def transform_data(self, data, theta, outsize: int, method=None, time=1.0):
        """ Combines the transform_grid and interpolate methods

        Args:
            data (numpy.ndarray or torch.Tensor): [n_batch, n_points, n_channels] tensor with input data.
            theta (numpy.ndarray or torch.Tensor): [n_batch, d] matrix
            outsize: number of points that is transformed and interpolated
            method (str): integration method, one of "*closed_form*", "*numeric*".
            time (float): integration time. Defaults to 1.0
        
        Returns:
            numpy.ndarray or torch.Tensor: [n_batch, outsize, n_channels] tensor, transformed and interpolated data
        """

        self._check_type(data)
        self._check_device(data)
        self._check_type(theta)
        self._check_device(theta)
        assert data.shape[0] == theta.shape[0], """Batch sizes should be the same on arguments data and theta"""
        grid = self.uniform_meshgrid(outsize)
        grid_t = self.transform_grid(grid, theta, method, time)
        data_t = self.interpolate(data, grid_t, outsize)
        return data_t

    def transform_data_ss(self, data, theta, outsize, method=None, time=1.0, N=0):
        """ Combines the transform_grid with scaling-squaring and interpolate methods

        Args:
            data (numpy.ndarray or torch.Tensor): [n_batch, n_points, n_channels] tensor with input data
            theta (numpy.ndarray or torch.Tensor): [n_batch, d] matrix
            outsize: number of points that is transformed (with scaling and squaring) and interpolated
            method (str): integration method, one of "*closed_form*", "*numeric*"
            time (float): integration time. Defaults to 1.0
            N (int): number of scaling iterations. Defaults to 0
        
        Returns:
            numpy.ndarray or torch.Tensor: [n_batch, outsize, n_channels] tensor, transformed and interpolated data
        """

        self._check_type(data)
        self._check_device(data)
        self._check_type(theta)
        self._check_device(theta)
        assert data.shape[0] == theta.shape[0], """Batch sizes should be the same on arguments data and theta"""
        grid = self.uniform_meshgrid(outsize)
        grid_t = self.transform_grid_ss(grid, theta, method, time, N)
        data_t = self.interpolate(data, grid_t, outsize)
        return data_t

    def calc_velocity(self, grid, theta):
        """ Calculates the velocity for each point in grid based on the theta parametrization

        Args:
            grid (numpy.ndarray or torch.Tensor): [n_batch, n_points] tensor with grid points that are used to interpolate the data
            theta (numpy.ndarray or torch.Tensor): [n_batch, d] matrix
        
        Returns:    
            numpy.ndarray or torch.Tensor: [n_points] vector with velocity values
        """
        self._check_type(grid)
        self._check_device(grid)
        self._check_type(theta)
        self._check_device(theta)
        v = self.backend.calc_velocity(grid, theta, self.params)
        return v

    def visualize_velocity(self, theta, n_points: int = None, fig: matplotlib.figure.Figure = None):
        """ Visualizes the vectorfield for a specific parametrization vector theta 

        Args:    
            theta (numpy.ndarray or torch.Tensor): [n_batch, d] matrix
            n_points (int): number of points to plot
            fig (matplotlib.figure.Figure): matplotlib figure handle
        
        Returns:
            matplotlib.axes.Axes: handle to lines plot
        """
        self._check_type(theta)
        if fig is None:
            fig = plt.figure()

        if n_points is None:
            n_points = self.params.nc + 1

        # Calculate vectorfield and convert to numpy
        grid = self.uniform_meshgrid(n_points)
        v = self.calc_velocity(grid, theta)

        grid = self.backend.tonumpy(grid)
        v = self.backend.tonumpy(v)

        # Plot
        ax = fig.add_subplot(1, 1, 1)
        ax.axhline(color="black", ls="dashed")
        alpha = max(0.01, 1 / np.sqrt(len(theta)))
        ax.plot(grid, v.T, color="blue", alpha=alpha)
        ax.plot(grid, v.T, marker="o", linestyle="None", color="orange", alpha=alpha)
        ax.set_xlim(self.params.xmin, self.params.xmax)
        ax.set_title("Velocity Field " + r"$v(x)$")
        ax.set_xlabel(r"$x$", rotation="horizontal")
        ax.set_ylabel(r"$v(x)$", rotation="horizontal")
        ax.grid(alpha=0.3)
        return ax

    def visualize_deformgrid(
        self, theta, method=None, time: float = 1.0, n_points: int = 100, fig: matplotlib.figure.Figure = None
    ):
        """ Visualizes the deformation for a specific parametrization vector theta 

        Args:    
            theta (numpy.ndarray or torch.Tensor): [n_batch, d] matrix
            method (str): integration method, one of "*closed_form*", "*numeric*"
            time (float): integration time. Defaults to 1.0
            n_points (int): number of points to plot
            fig (matplotlib.figure.Figure): matplotlib figure handle
        
        Returns:
            matplotlib.axes.Axes: handle to lines plot
        """

        if fig is None:
            fig = plt.figure()

        # Transform grid and convert to numpy
        grid = self.uniform_meshgrid(n_points)
        grid_t = self.transform_grid(grid, theta, method, time)
        grid = self.backend.tonumpy(grid)
        grid_t = self.backend.tonumpy(grid_t)

        # Plot
        ax = fig.add_subplot(1, 1, 1)
        ax.axline((0, 0), (1, 1), color="black", ls="dashed")
        alpha = max(0.01, 1 / np.sqrt(len(theta)))
        ax.plot(grid, grid_t.T, c="blue", alpha=alpha)
        ax.set_title("Grid Deformation " + r"$\phi(x,t)$")
        ax.set_xlabel(r"$x$", rotation="horizontal")
        ax.set_ylabel(r"$\phi(x,t)$", rotation="horizontal")
        ax.grid(alpha=0.3)

        return ax

    def visualize_tesselation(self, n_points: int = 50, fig: matplotlib.figure.Figure = None):
        """ Visualizes the tesselation for a specific parametrization vector theta 

        Args:    
            theta (numpy.ndarray or torch.Tensor): [n_batch, d] matrix
            n_points (int): number of points to plot
            fig (matplotlib.figure.Figure): matplotlib figure handle
        
        Returns:
            matplotlib.axes.Axes: handle to tesselation plot
        """

        if fig is None:
            fig = plt.figure()

        grid = self.uniform_meshgrid(n_points)

        # Find cellindex and convert to numpy
        idx = self.backend.get_cell(grid, self.params)
        idx = self.backend.tonumpy(idx)
        grid = self.backend.tonumpy(grid)

        # Plot
        ax = fig.add_subplot(1, 1, 1)
        plot = ax.scatter(grid.flatten(), idx, c=idx)
        ax.set_title("Tesselation " + r"$N_\mathcal{P}=$" + str(self.params.nc))
        ax.set_xlabel(r"$x$")
        ax.set_ylabel("Cell Index")
        ax.set_yticks(np.arange(np.max(idx) + 1))
        ax.grid(alpha=0.3)
        return plot

    def visualize_gradient(
        self, theta, method=None, time: float = 1.0, n_points: int = 100, fig: matplotlib.figure.Figure = None
    ):
        """ Visualizes the gradient for a specific parametrization vector theta 

        Args:    
            theta (numpy.ndarray or torch.Tensor): [n_batch, d] matrix
            method (str): integration method, one of "*closed_form*", "*numeric*"
            time (float): integration time. Defaults to 1.0
            n_points (int): number of points to plot
            fig (matplotlib.figure.Figure): matplotlib figure handle
        
        Returns:
            matplotlib.axes.Axes: handle to gradient plot
        """
        if fig is None:
            fig = plt.figure()

        # Gradient grid and convert to numpy
        grid = self.uniform_meshgrid(n_points)
        grad = self.gradient_grid(grid, theta, method, time)
        grid = self.backend.tonumpy(grid)
        grad = self.backend.tonumpy(grad)

        # Plot only gradient
        ax = fig.add_subplot(1, 1, 1)
        ax.axhline(color="black", ls="dashed")
        for g in grad:
            ax.plot(grid, g)
        ax.set_title("Gradient " + r"$\dfrac{\partial \phi(x,t)}{\partial \theta}$")
        ax.set_xlabel(r"$x$", rotation="horizontal")
        ax.set_ylabel(
            r"$\dfrac{\partial \phi(x,t)}{\partial \theta}$", rotation="horizontal", labelpad=10,
        )
        ax.grid(alpha=0.3)
        return ax

    def visualize_deformdata(
        self, data, theta, method=None, outsize: int = 100, target=None, fig: matplotlib.figure.Figure = None
    ):
        """ Visualizes the transformed data for a specific parametrization vector theta 

        Args:    
            theta (numpy.ndarray or torch.Tensor): [n_batch, d] matrix
            method (str): integration method, one of "*closed_form*", "*numeric*"
            outsize (int): number of points that is transformed and interpolated
            target (numpy.ndarray or torch.Tensor): [n_batch, n_points, n_channels] tensor with target data.
            time (float): integration time. Defaults to 1.0
            n_points (int): number of points to plot
            fig (matplotlib.figure.Figure): matplotlib figure handle
        
        Returns:
            matplotlib.axes.Axes: handle to gradient plot
        """

        if fig is None:
            fig = plt.figure()

        # Transform grid and convert to numpy
        data_t = self.transform_data(data, theta, outsize, method)
        data_t = self.backend.tonumpy(data_t)

        batch_size, width, channels = data.shape
        x = np.linspace(0, 1, width)
        xt = np.linspace(0, 1, outsize)
        alpha = max(0.01, 1 / np.sqrt(batch_size))

        # Plot
        plt.suptitle("Data Deformation with " + r"$\phi(x,t)$")
        for i in range(channels):
            ax = fig.add_subplot(channels, 1, i + 1)
            ax.plot(x, data[:, :, i].T, c="red", ls="dashed", alpha=alpha, label="Data")
            if target is not None:
                ax.plot(
                    xt, target[:, :, i].T, c="green", ls="dashdot", alpha=alpha, label="Target",
                )
            ax.plot(xt, data_t[:, :, i].T, alpha=alpha, label="Transformed")
            ax.set_title("Ch " + str(i), rotation=-90, loc="right", y=0.5, ha="left", va="center")
            ax.set_ylabel(r"$x'$", rotation="horizontal")
            ax.grid(alpha=0.3)
            if i + 1 < channels:
                ax.set_xticklabels([])
            else:
                ax.set_xlabel(r"$x$", rotation="horizontal")

        return ax

    def _check_input(self, tess_size: int, backend: str, device: str, zero_boundary: bool, basis: str):
        """ Checks the input to the class

            tess_size (int): tesselation size, with the number of vertices.
            backend (str, optional): computational backend to use. Choose between "*numpy*", or "*pytorch*". Defaults to "*numpy*".
            device (str, optional): device to use. Choose between "*cpu*" or "*gpu*". For the numpy backend only the "*cpu*" option is valid. Defaults to "*cpu*".
            zero_boundary (bool, optional): determines if the velocity at the boundary is zero 
            basis (str, optional): constrain basis to use. Choose between "*rref*", "*svd*", or "*sparse*". Defaults to "*rref*".
            
        """

        assert tess_size > 1, """tess size must be > 1"""
        assert backend in ["numpy", "pytorch",], """Unknown backend, choose between 'numpy' or 'pytorch' """
        assert device in ["cpu", "gpu",], """Unknown device, choose between 'cpu' or 'gpu' """
        if backend == "numpy":
            assert device == "cpu", """Cannot use gpu with numpy backend """
        assert type(zero_boundary) == bool, """Argument zero_boundary must be True or False"""
        assert basis in [
            "svd",
            "rref",
            "sparse",
            "qr",
        ], """Unknown basis, choose between 'svd', 'rref' 'qr', or 'sparse' """

    def _check_type(self, x):
        """ Asserts that the type of x is compatible with the class i.e
                numpy backend expects np.array
                pytorch backend expects torch.tensor
        """
        assert isinstance(x, self.backend.backend_type()), """ Input has type {0} but expected type {1} """.format(
            type(x), self.backend.backend_type()
        )

    def _check_device(self, x):
        """ Asserts that x is on the same device (cpu or gpu) as the class """
        assert self.backend.check_device(
            x, self.device
        ), """Input is placed on 
            device {0} but the class expects it to be on device {1}""".format(
            str(x.device), self.device
        )

    def __repr__(self):
        output = """
        CPAB transformer class. 
            Parameters:
                Tesselation size:           {0}
                Theta size:                 {1}
                Domain lower bound:         {2}
                Domain upper bound:         {3}
                Zero Boundary:              {4}
            Backend:                        {5}
        """.format(
            self.params.nc,
            self.params.d,
            self.params.xmin,
            self.params.xmax,
            self.params.zero_boundary,
            self.backend_name,
        )
        return output



# %%
import numpy as np
from .interpolation import interpolate, interpolate_grid as interpolate_grid_wrapper
from .transformer import (
    integrate_numeric,
    integrate_closed_form,
    derivative_numeric,
    derivative_closed_form,
    derivative_space_numeric,
    derivative_space_closed_form
)
from .transformer import get_cell, get_velocity, batch_effect
from ...core.utility import methods

# %%
def assert_version():
    numbers = np.__version__.split(".")
    version = float(numbers[0] + "." + numbers[1])
    assert (
        version >= 1.14
    ), """ You are using a older installation of numpy, please install 1.15.
            or newer """


# %%
half = np.float16
single = np.float32
double = np.float64


def to(x, dtype=np.float32, device=None):
    return np.array(x)


# %%
def tonumpy(x):
    return x


# %%
def check_device(x, device_name):
    return True  # always return true, because device can only be cpu


# %%
def backend_type():
    return np.ndarray


# %%
def sample_transformation(d, n_sample=1, mean=None, cov=None, device="cpu"):
    mean = np.zeros(d, dtype=np.float32) if mean is None else mean
    cov = np.eye(d, dtype=np.float32) if cov is None else cov
    samples = np.random.multivariate_normal(mean, cov, size=n_sample)
    return samples


# %%
def identity(d, n_sample=1, epsilon=0, device="cpu"):
    assert epsilon >= 0, "epsilon need to be larger than or 0"
    return np.zeros((n_sample, d), dtype=np.float32) + epsilon


# %%
def uniform_meshgrid(xmin, xmax, n_points, device="cpu"):
    return np.linspace(xmin, xmax, n_points)


# %%
def calc_velocity(grid, theta, params):
    grid = batch_effect(grid, theta)
    v = get_velocity(grid, theta, params)
    return v.reshape(theta.shape[0], -1)


# %%
def exp(*args, **kwargs):
    return np.exp(*args, **kwargs)


def linspace(*args, **kwargs):
    return np.linspace(*args, **kwargs)


def meshgrid(*args, **kwargs):
    return np.meshgrid(*args, **kwargs)


def matmul(*args, **kwargs):
    return np.matmul(*args, **kwargs)


def max(*args, **kwargs):
    return np.max(*args, **kwargs)


def ones(*args, **kwargs):
    return np.ones(*args, **kwargs)


def pdist(c):
    x, y = np.meshgrid(c, c)
    return np.abs(x - y)


# %%


def transformer(grid, theta, params, method=None, time=1.0):
    methods.check(method)
    method = methods.default(method)

    if method == methods.closed_form:
        return integrate_closed_form(grid, theta, params, time)
    elif method == methods.numeric:
        return integrate_numeric(grid, theta, params, time)


# %%
def gradient(grid, theta, params, method=None, time=1.0):
    methods.check(method)
    method = methods.default(method)

    if method == methods.closed_form:
        phi, der = derivative_closed_form(grid, theta, params, time)
        return der
    elif method == methods.numeric:
        h = 1e-2
        phi, der = derivative_numeric(grid, theta, params, time, h)
        return der

# %%
def gradient_space(grid, theta, params, method=None, time=1.0):
    methods.check(method)
    method = methods.default(method)

    if method == methods.closed_form:
        phi, der = derivative_space_closed_form(grid, theta, params, time)
        return der
    elif method == methods.numeric:
        h = 1e-2
        phi, der = derivative_space_numeric(grid, theta, params, time, h)
        return der

# %%
def interpolate_grid(transformed_grid, params):
    return interpolate_grid_wrapper(transformed_grid)


# %%
import torch
from torch.utils.cpp_extension import load
from ...core.utility import get_dir, methods

# %% NOT COMPILED HELPER
class _notcompiled:
    # Small class, with structure similar to the compiled modules we can default
    # to. The class will never be called but the program can compile at run time
    def __init__(self):
        def f(*args):
            return None

        self.forward = f
        self.backward = f


# %% COMPILE AND LOAD

# TODO: change slow and fast for c++ / python or compiled / not compiled
_dir = get_dir(__file__)
_verbose = True

# Jit compile cpu source
_cpu_success = False
cpab_cpu = _notcompiled()
try:
    cpab_cpu = load(
        name="cpab_cpu",
        sources=[_dir + "/transformer.cpp", _dir + "/../../core/cpab_ops.cpp"],
        # extra_cflags=["-O0", "-g"],
        extra_cflags=["-Ofast", "-ffast-math", "-funsafe-math-optimizations",],  # , "-msse4.2"
        verbose=_verbose,
    )
    _cpu_success = True
    if _verbose:
        print(70 * "=")
        print("succesfully compiled cpu source")
        print(70 * "=")
except Exception as e:
    cpab_cpu = _notcompiled()
    _cpu_success = False
    if _verbose:
        print(70 * "=")
        print("Unsuccesfully compiled cpu source")
        print("Error was: ")
        print(e)
        print(70 * "=")


# Jit compile gpu source
_gpu_success = False
cpab_gpu = _notcompiled()
try:
    cpab_gpu = load(
        name="cpab_gpu",
        sources=[_dir + "/transformer_cuda.cpp", _dir + "/transformer_cuda.cu", _dir + "/../../core/cpab_ops.cu",],
        # extra_cflags=["-O0", "-g"],
        extra_cflags=["-Ofast", "-ffast-math", "-funsafe-math-optimizations"],
        verbose=_verbose,
        with_cuda=True,
    )
    _gpu_success = True
    if _verbose:
        print(70 * "=")
        print("succesfully compiled gpu source")
        print(70 * "=")
except Exception as e:
    cpab_gpu = _notcompiled()
    _gpu_success = False
    if _verbose:
        print(70 * "=")
        print("Unsuccesfully compiled gpu source")
        print("Error was: ")
        print(e)


_verbose = False
# %% GET CELL

from .transformer_slow import get_cell as get_cell_slow


def get_cell(grid, params):
    if grid.is_cuda:
        if not params.use_slow and _gpu_success:
            if _verbose:
                print("using fast gpu implementation")
            return cpab_gpu.get_cell(grid, params.xmin, params.xmax, params.nc)
        else:
            if _verbose:
                print("using slow gpu implementation")
            return get_cell_slow(grid, params)
    else:
        if not params.use_slow and _cpu_success:
            if _verbose:
                print("using fast cpu implementation")
            return cpab_cpu.get_cell(grid, params.xmin, params.xmax, params.nc)
        else:
            if _verbose:
                print("using slow cpu implementation")
            return get_cell_slow(grid, params)


# %% CALC VELOCITY

from .transformer_slow import calc_velocity as calc_velocity_slow


def calc_velocity(grid, theta, params):
    if grid.is_cuda and theta.is_cuda:
        if not params.use_slow and _gpu_success:
            if _verbose:
                print("using fast gpu implementation")
            return Velocity_gpu.apply(grid, theta, params)
        else:
            if _verbose:
                print("using slow gpu implementation")
            return calc_velocity_slow(grid, theta, params)
    else:
        if not params.use_slow and _cpu_success:
            if _verbose:
                print("using fast cpu implementation")
            return Velocity_cpu.apply(grid, theta, params)
        else:
            if _verbose:
                print("using slow cpu implementation")
            return calc_velocity_slow(grid, theta, params)


#%% VELOCITY: CPU


class Velocity_cpu(torch.autograd.Function):
    @staticmethod
    def forward(ctx, grid, theta, params):
        v = cpab_cpu.get_velocity(grid, theta, params.B, params.xmin, params.xmax, params.nc)
        ctx.params = params
        ctx.save_for_backward(grid, theta)
        return v

    @staticmethod
    @torch.autograd.function.once_differentiable
    def backward(ctx, grad_output):  # grad_output [n_batch, n_points]
        grid, theta = ctx.saved_tensors
        params = ctx.params
        
        dv_dtheta = cpab_cpu.derivative_velocity_dtheta(grid, theta, params.B, params.xmin, params.xmax, params.nc).t()
        grad_theta = grad_output.matmul(dv_dtheta)

        dv_dx = cpab_cpu.derivative_velocity_dx(grid, theta, params.B, params.xmin, params.xmax, params.nc)
        grad_x = grad_output.mul(dv_dx)

        return grad_x, grad_theta, None  # [n_batch, n_points] # [n_batch, d]


#%% VELOCITY: GPU


class Velocity_gpu(torch.autograd.Function):
    @staticmethod
    def forward(ctx, grid, theta, params):
        v = cpab_gpu.get_velocity(grid, theta, params.B, params.xmin, params.xmax, params.nc)
        ctx.params = params
        ctx.save_for_backward(grid, theta)
        return v

    @staticmethod
    @torch.autograd.function.once_differentiable
    def backward(ctx, grad_output):  # grad_output [n_batch, n_points]
        grid, theta = ctx.saved_tensors
        params = ctx.params
        
        dv_dtheta = cpab_gpu.derivative_velocity_dtheta(grid, theta, params.B, params.xmin, params.xmax, params.nc).t()
        grad_theta = grad_output.matmul(dv_dtheta)

        dv_dx = cpab_gpu.derivative_velocity_dx(grid, theta, params.B, params.xmin, params.xmax, params.nc)
        grad_x = grad_output.mul(dv_dx)

        return grad_x, grad_theta, None  # [n_batch, n_points] # [n_batch, d]


# %% GRADIENT


def gradient(grid, theta, params, method=None, time=1.0):
    if grid.is_cuda and theta.is_cuda:
        if not params.use_slow and _gpu_success:
            if _verbose:
                print("using fast gpu implementation")
            return gradient_fast_gpu(grid, theta, params, method, time)
        else:
            if _verbose:
                print("using slow gpu implementation")
            return gradient_slow(grid, theta, params, method, time)
    else:
        if not params.use_slow and _cpu_success:
            if _verbose:
                print("using fast cpu implementation")
            return gradient_fast_cpu(grid, theta, params, method, time)
        else:
            if _verbose:
                print("using slow cpu implementation")
            return gradient_slow(grid, theta, params, method, time)


# %% GRADIENT: SLOW / CPU + GPU
from .transformer_slow import derivative_numeric, derivative_closed_form


def gradient_slow(grid, theta, params, method=None, time=1.0):
    methods.check(method)
    method = methods.default(method)

    if method == methods.closed_form:
        phi, der = derivative_closed_form(grid, theta, params, time)
        return der
    elif method == methods.numeric:
        h = 1e-2
        phi, der = derivative_numeric(grid, theta, params, time, h)
        return der


# %% GRADIENT: FAST / CPU
def gradient_fast_cpu(grid, theta, params, method=None, time=1.0):
    methods.check(method)
    method = methods.default(method)

    if method == methods.closed_form:
        return cpab_cpu.derivative_closed_form(grid, theta, time, params.B, params.xmin, params.xmax, params.nc)
    elif method == methods.numeric:
        h = 1e-2
        return cpab_cpu.derivative_numeric(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc, params.nSteps1, params.nSteps2, h,
        )


# %% GRADIENT: FAST / GPU
def gradient_fast_gpu(grid, theta, params, method=None, time=1.0):
    methods.check(method)
    method = methods.default(method)

    if method == methods.closed_form:
        return cpab_gpu.derivative_closed_form(grid, theta, time, params.B, params.xmin, params.xmax, params.nc)
    elif method == methods.numeric:
        h = 1e-2
        return cpab_gpu.derivative_numeric(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc, params.nSteps1, params.nSteps2, h,
        )


# %% TRANSFORMER
def transformer(grid, theta, params, method=None, time=1.0):
    if grid.is_cuda and theta.is_cuda:
        if not params.use_slow and _gpu_success:
            if _verbose:
                print("using fast gpu implementation")
            return transformer_fast_gpu(grid, theta, params, method, time)
        else:
            if _verbose:
                print("using slow gpu implementation")
            return transformer_slow(grid, theta, params, method, time)
    else:
        if not params.use_slow and _cpu_success:
            if _verbose:
                print("using fast cpu implementation")
            return transformer_fast_cpu(grid, theta, params, method, time)
        else:
            if _verbose:
                print("using slow cpu implementation")
            return transformer_slow(grid, theta, params, method, time)


# %% TRANSFORMER: SLOW / CPU + GPU
from .transformer_slow import integrate_closed_form, integrate_numeric


def transformer_slow(grid, theta, params, method=None, time=1.0):
    methods.check(method)
    method = methods.default(method)

    # TODO: testing benchmark
    if method == methods.closed_form:
        # return integrate_closed_form(grid, theta, params)
        return Transformer_slow_closed_form.apply(grid, theta, params, time)
    elif method == methods.numeric:
        # return integrate_numeric(grid, theta, params)
        return Transformer_slow_numeric.apply(grid, theta, params, time)


#%% TRANSFORMER: SLOW / CPU + GPU / CLOSED-FORM
from .transformer_slow import integrate_closed_form_trace, derivative_closed_form_trace, derivative_space_closed_form


class Transformer_slow_closed_form(torch.autograd.Function):
    @staticmethod
    def forward(ctx, grid, theta, params, time=1.0):
        ctx.params = params
        ctx.time = time

        output = integrate_closed_form_trace(grid, theta, params, time)
        n_batch = theta.shape[0]
        grid_t = output[:, 0].reshape((n_batch, -1)).contiguous()
        
        ctx.save_for_backward(output, grid, theta)
        return grid_t

    @staticmethod
    @torch.autograd.function.once_differentiable
    def backward(ctx, grad_output):  # grad_output [n_batch, n_points]
        output, grid, theta = ctx.saved_tensors
        params = ctx.params
        time = ctx.time

        dphi_dtheta = derivative_closed_form_trace(output, grid, theta, params)  # [n_batch, n_points, d]
        # NOTE: we have to permute the gradient in order to do the element-wise product
        # Then, the gradient must be summarized for all grid points
        grad_theta = grad_output.mul(dphi_dtheta.permute(2, 0, 1)).sum(dim=(2)).t()

        _, dphi_dx = derivative_space_closed_form(grid, theta, params, time=1.0)  # [n_batch, n_points]
        grad_x = grad_output.mul(dphi_dx)

        return grad_x, grad_theta, None, None  # [n_batch, n_points] # [n_batch, d]


#%% TRANSFORMER: SLOW / CPU + GPU / NUMERIC
from .transformer_slow import integrate_numeric, derivative_numeric_trace, derivative_space_numeric


class Transformer_slow_numeric(torch.autograd.Function):
    @staticmethod
    def forward(ctx, grid, theta, params, time=1.0):
        ctx.params = params
        ctx.time = time
        
        grid_t = integrate_numeric(grid, theta, params, time)
        
        ctx.save_for_backward(grid_t, grid, theta)
        return grid_t

    @staticmethod
    @torch.autograd.function.once_differentiable
    def backward(ctx, grad_output):  # grad_output [n_batch, n_points]
        grid_t, grid, theta = ctx.saved_tensors
        params = ctx.params
        time = ctx.time

        h = 1e-2
        dphi_dtheta = derivative_numeric_trace(grid_t, grid, theta, params, time, h)
        grad_theta = grad_output.mul(dphi_dtheta.permute(2, 0, 1)).sum(dim=(2)).t()

        _, dphi_dx = derivative_space_numeric(grid, theta, params, time, h)
        grad_x = grad_output.mul(dphi_dx)

        return grad_x, grad_theta, None, None  # [n_batch, n_points] # [n_batch, d]


# %% TRANSFORMER: FAST / CPU
def transformer_fast_cpu(grid, theta, params, method=None, time=1.0):
    methods.check(method)
    method = methods.default(method)

    if method == methods.closed_form:
        return Transformer_fast_cpu_closed_form.apply(grid, theta, params, time)
    elif method == methods.numeric:
        return Transformer_fast_cpu_numeric.apply(grid, theta, params, time)


# %% TRANSFORMER: FAST / CPU / CLOSED-FORM
class Transformer_fast_cpu_closed_form(torch.autograd.Function):
    @staticmethod
    def forward(ctx, grid, theta, params, time=1.0):
        ctx.params = params
        ctx.time = time

        output = cpab_cpu.integrate_closed_form_trace(grid, theta, time, params.B, params.xmin, params.xmax, params.nc)
        grid_t = output[:, :, 0].contiguous()
        
        ctx.save_for_backward(output, grid, theta)
        return grid_t

    @staticmethod
    @torch.autograd.function.once_differentiable
    def backward(ctx, grad_output):  # grad_output [n_batch, n_points]
        output, grid, theta = ctx.saved_tensors
        params = ctx.params
        time = ctx.time

        dphi_dtheta = cpab_cpu.derivative_closed_form_trace(
            output, grid, theta, params.B, params.xmin, params.xmax, params.nc
        )  # [n_batch, n_points, d]
        grad_theta = grad_output.mul(dphi_dtheta.permute(2, 0, 1)).sum(dim=(2)).t()

        dphi_dx = cpab_cpu.derivative_space_closed_form(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc)
        grad_x = grad_output.mul(dphi_dx)

        return grad_x, grad_theta, None, None  # [n_batch, n_points] # [n_batch, d]


# %% TRANSFORMER: FAST / CPU / NUMERIC
class Transformer_fast_cpu_numeric(torch.autograd.Function):
    @staticmethod
    def forward(ctx, grid, theta, params, time=1.0):
        ctx.params = params
        ctx.time = time
        grid_t = cpab_cpu.integrate_numeric(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc, params.nSteps1, params.nSteps2,
        )
        ctx.save_for_backward(grid_t, grid, theta)
        return grid_t

    @staticmethod
    @torch.autograd.function.once_differentiable
    def backward(ctx, grad_output):  # grad_output [n_batch, n_points]
        params = ctx.params
        time = ctx.time
        grid_t, grid, theta = ctx.saved_tensors

        h = 1e-2
        dphi_dtheta = cpab_cpu.derivative_numeric_trace(
            grid_t, grid, theta, time, params.B, params.xmin, params.xmax, params.nc, params.nSteps1, params.nSteps2, h,
        )
        grad_theta = grad_output.mul(dphi_dtheta.permute(2, 0, 1)).sum(dim=(2)).t()

        dphi_dx = cpab_cpu.derivative_space_numeric(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc, params.nSteps1, params.nSteps2, h,
        )
        grad_x = grad_output.mul(dphi_dx)

        return grad_x, grad_theta, None, None  # [n_batch, n_points] # [n_batch, d]


# %% TRANSFORMER: FAST / GPU
def transformer_fast_gpu(grid, theta, params, method=None, time=1.0):
    methods.check(method)
    method = methods.default(method)

    if method == methods.closed_form:
        return Transformer_fast_gpu_closed_form.apply(grid, theta, params, time)
    elif method == methods.numeric:
        return Transformer_fast_gpu_numeric.apply(grid, theta, params, time)


# %% TRANSFORMER: FAST / GPU / CLOSED-FORM
class Transformer_fast_gpu_closed_form(torch.autograd.Function):
    @staticmethod
    def forward(ctx, grid, theta, params, time=1.0):
        ctx.params = params
        ctx.time = time
        output = cpab_gpu.integrate_closed_form_trace(grid, theta, time, params.B, params.xmin, params.xmax, params.nc)
        grid_t = output[:, :, 0].contiguous()
        ctx.save_for_backward(output, grid, theta)
        return grid_t

    @staticmethod
    @torch.autograd.function.once_differentiable
    def backward(ctx, grad_output):  # grad [n_batch, n_points]
        output, grid, theta = ctx.saved_tensors
        params = ctx.params
        time = ctx.time

        dphi_dtheta = cpab_gpu.derivative_closed_form_trace(
            output, grid, theta, params.B, params.xmin, params.xmax, params.nc
        )  # [n_batch, n_points, d]
        grad_theta = grad_output.mul(dphi_dtheta.permute(2, 0, 1)).sum(dim=(2)).t()

        dphi_dx = cpab_gpu.derivative_space_closed_form(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc)
        grad_x = grad_output.mul(dphi_dx)

        return grad_x, grad_theta, None, None  # [n_batch, n_points] # [n_batch, d]


# %% TRANSFORMER: FAST / GPU / NUMERIC
class Transformer_fast_gpu_numeric(torch.autograd.Function):
    # Function that connects the forward pass to the backward pass
    @staticmethod
    def forward(ctx, grid, theta, params, time=1.0):
        ctx.params = params
        ctx.time = time
        grid_t = cpab_gpu.integrate_numeric(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc, params.nSteps1, params.nSteps2,
        )
        ctx.save_for_backward(grid_t, grid, theta)
        return grid_t

    @staticmethod
    @torch.autograd.function.once_differentiable
    def backward(ctx, grad_output):  # grad_output [n_batch, n_points]
        params = ctx.params
        time = ctx.time
        grid_t, grid, theta = ctx.saved_tensors

        h = 1e-2
        dphi_dtheta = cpab_gpu.derivative_numeric_trace(
            grid_t, grid, theta, time, params.B, params.xmin, params.xmax, params.nc, params.nSteps1, params.nSteps2, h,
        )
        grad_theta = grad_output.mul(dphi_dtheta.permute(2, 0, 1)).sum(dim=(2)).t()

        dphi_dx = cpab_gpu.derivative_space_numeric(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc, params.nSteps1, params.nSteps2, h,
        )
        grad_x = grad_output.mul(dphi_dx)

        return grad_x, grad_theta, None, None  # [n_batch, n_points] # [n_batch, d]


# %% INTERPOLATE
from .interpolation import interpolate_grid as interpolate_grid_slow


def interpolate_grid(transformed_grid, params):
    if transformed_grid.is_cuda:
        if not params.use_slow and _gpu_success:
            if _verbose:
                print("using fast gpu implementation")
            return Interpolate_fast_gpu.apply(transformed_grid)
        else:
            if _verbose:
                print("using slow gpu implementation")
            return interpolate_grid_slow(transformed_grid)
    else:
        if not params.use_slow and _cpu_success:
            if _verbose:
                print("using fast cpu implementation")
            return Interpolate_fast_cpu.apply(transformed_grid)
        else:
            if _verbose:
                print("using slow cpu implementation")
            return interpolate_grid_slow(transformed_grid)


# %% INTERPOLATE: FAST / CPU
class Interpolate_fast_cpu(torch.autograd.Function):
    @staticmethod
    def forward(ctx, transformed_grid):
        ctx.save_for_backward(transformed_grid)
        return cpab_cpu.interpolate_grid_forward(transformed_grid)

    @staticmethod
    @torch.autograd.function.once_differentiable
    def backward(ctx, grad_output):  # grad [n_batch, n_points]
        (transformed_grid,) = ctx.saved_tensors
        return cpab_cpu.interpolate_grid_backward(grad_output, transformed_grid)


# %% INTERPOLATE: FAST / GPU
class Interpolate_fast_gpu(torch.autograd.Function):
    @staticmethod
    def forward(ctx, transformed_grid):
        ctx.save_for_backward(transformed_grid)
        return cpab_gpu.interpolate_grid_forward(transformed_grid)

    @staticmethod
    @torch.autograd.function.once_differentiable
    def backward(ctx, grad_output):  # grad [n_batch, n_points]
        (transformed_grid,) = ctx.saved_tensors
        return cpab_gpu.interpolate_grid_backward(grad_output, transformed_grid)

# %% GRADIENT SPACE


def gradient_space(grid, theta, params, method=None, time=1.0):
    if grid.is_cuda and theta.is_cuda:
        if not params.use_slow and _gpu_success:
            if _verbose:
                print("using fast gpu implementation")
            return gradient_space_fast_gpu(grid, theta, params, method, time)
        else:
            if _verbose:
                print("using slow gpu implementation")
            return gradient_space_slow(grid, theta, params, method, time)
    else:
        if not params.use_slow and _cpu_success:
            if _verbose:
                print("using fast cpu implementation")
            return gradient_space_fast_cpu(grid, theta, params, method, time)
        else:
            if _verbose:
                print("using slow cpu implementation")
            return gradient_space_slow(grid, theta, params, method, time)


# %% GRADIENT SPACE: SLOW / CPU + GPU
from .transformer_slow import derivative_space_numeric, derivative_space_closed_form


def gradient_space_slow(grid, theta, params, method=None, time=1.0):
    methods.check(method)
    method = methods.default(method)

    if method == methods.closed_form:
        phi, der = derivative_space_closed_form(grid, theta, params, time)
        return der
    elif method == methods.numeric:
        h = 1e-2
        phi, der = derivative_space_numeric(grid, theta, params, time, h)
        return der



# %% GRADIENT SPACE: FAST / CPU
def gradient_space_fast_cpu(grid, theta, params, method=None, time=1.0):
    methods.check(method)
    method = methods.default(method)

    if method == methods.closed_form:
        return GradientSpace_fast_cpu_closed_form.apply(grid, theta, params, time)
    elif method == methods.numeric:
        return GradientSpace_fast_cpu_numeric.apply(grid, theta, params, time)

# %% GRADIENT SPACE: FAST / CPU / NUMERIC
class GradientSpace_fast_cpu_numeric(torch.autograd.Function):
    @staticmethod
    def forward(ctx, grid, theta, params, time=1.0):
        ctx.params = params
        ctx.time = time
        
        h = 1e-2
        # dphi_dx = cpab_cpu.derivative_space_numeric(
        #     grid, theta, time, params.B, params.xmin, params.xmax, params.nc, params.nSteps1, params.nSteps2, h,
        # )
        dphi_dx = cpab_cpu.derivative_space_closed_form(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc
        )
        ctx.save_for_backward(dphi_dx, grid, theta)
        return dphi_dx

    @staticmethod
    @torch.autograd.function.once_differentiable
    def backward(ctx, grad_output):  # grad_output [n_batch, n_points]
        params = ctx.params
        time = ctx.time
        dphi_dx, grid, theta = ctx.saved_tensors

        h = 1e-3
        dphi_dx_dtheta = cpab_cpu.derivative_space_numeric_dtheta(
            dphi_dx, grid, theta, time, params.B, params.xmin, params.xmax, params.nc, params.nSteps1, params.nSteps2, h)
        grad_theta = grad_output.mul(dphi_dx_dtheta).sum(dim=(2)).t()

        dphi_dx_dx = cpab_cpu.derivative_space_numeric_dx(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc, params.nSteps1, params.nSteps2, h)
        grad_x = grad_output.mul(dphi_dx_dx)
        
        return grad_x, grad_theta, None, None  # [n_batch, n_points] # [n_batch, d]

# %% GRADIENT SPACE: FAST / CPU / CLOSED-FORM
class GradientSpace_fast_cpu_closed_form(torch.autograd.Function):
    @staticmethod
    def forward(ctx, grid, theta, params, time=1.0):
        ctx.params = params
        ctx.time = time
        
        dphi_dx = cpab_cpu.derivative_space_closed_form(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc)
        ctx.save_for_backward(grid, theta)
        return dphi_dx

    @staticmethod
    @torch.autograd.function.once_differentiable
    def backward(ctx, grad_output):  # grad_output [n_batch, n_points]
        params = ctx.params
        time = ctx.time
        grid, theta = ctx.saved_tensors

        dphi_dx_dtheta = cpab_cpu.derivative_space_closed_form_dtheta(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc)
        grad_theta = grad_output.mul(dphi_dx_dtheta.permute(2,0,1)).sum(dim=(2)).t()

        dphi_dx_dx = cpab_cpu.derivative_space_closed_form_dx(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc)
        grad_x = grad_output.mul(dphi_dx_dx)

        return grad_x, grad_theta, None, None  # [n_batch, n_points] # [n_batch, d]

# %% GRADIENT SPACE: FAST / GPU
def gradient_space_fast_gpu(grid, theta, params, method=None, time=1.0):
    methods.check(method)
    method = methods.default(method)

    if method == methods.closed_form:
        return GradientSpace_fast_gpu_closed_form.apply(grid, theta, params, time)
    elif method == methods.numeric:
        return GradientSpace_fast_gpu_numeric.apply(grid, theta, params, time)

# %% GRADIENT SPACE: FAST / GPU / NUMERIC
class GradientSpace_fast_gpu_numeric(torch.autograd.Function):
    @staticmethod
    def forward(ctx, grid, theta, params, time=1.0):
        ctx.params = params
        ctx.time = time
        
        h = 1e-2
        # dphi_dx = cpab_gpu.derivative_space_numeric(
        #     grid, theta, time, params.B, params.xmin, params.xmax, params.nc, params.nSteps1, params.nSteps2, h,
        # )
        dphi_dx = cpab_gpu.derivative_space_closed_form(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc
        )
        ctx.save_for_backward(dphi_dx, grid, theta)
        return dphi_dx

    @staticmethod
    @torch.autograd.function.once_differentiable
    def backward(ctx, grad_output):  # grad_output [n_batch, n_points]
        params = ctx.params
        time = ctx.time
        dphi_dx, grid, theta = ctx.saved_tensors

        h = 1e-3
        dphi_dx_dtheta = cpab_gpu.derivative_space_numeric_dtheta(
            dphi_dx, grid, theta, time, params.B, params.xmin, params.xmax, params.nc, params.nSteps1, params.nSteps2, h,
        )
        grad_theta = grad_output.mul(dphi_dx_dtheta).sum(dim=(2)).t()

        dphi_dx_dx = cpab_gpu.derivative_space_numeric_dx(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc, params.nSteps1, params.nSteps2, h,
        )
        grad_x = grad_output.mul(dphi_dx_dx)
        
        return grad_x, grad_theta, None, None  # [n_batch, n_points] # [n_batch, d]

# %% GRADIENT SPACE: FAST / GPU / CLOSED-FORM
class GradientSpace_fast_gpu_closed_form(torch.autograd.Function):
    @staticmethod
    def forward(ctx, grid, theta, params, time=1.0):
        ctx.params = params
        ctx.time = time
        
        dphi_dx = cpab_gpu.derivative_space_closed_form(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc)
        ctx.save_for_backward(grid, theta)
        return dphi_dx

    @staticmethod
    @torch.autograd.function.once_differentiable
    def backward(ctx, grad_output):  # grad_output [n_batch, n_points]
        params = ctx.params
        time = ctx.time
        grid, theta = ctx.saved_tensors

        dphi_dx_dtheta = cpab_gpu.derivative_space_closed_form_dtheta(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc)
        grad_theta = grad_output.mul(dphi_dx_dtheta.permute(2,0,1)).sum(dim=(2)).t()

        dphi_dx_dx = cpab_gpu.derivative_space_closed_form_dx(
            grid, theta, time, params.B, params.xmin, params.xmax, params.nc)
        grad_x = grad_output.mul(dphi_dx_dx)

        return grad_x, grad_theta, None, None  # [n_batch, n_points] # [n_batch, d]



import torch


def interpolate(data, grid, outsize):
    # Problem size
    n_batch, width, n_channels = data.shape

    # Extract points
    x = grid.flatten()

    # Scale to domain
    x = x * (width - 1)

    # Do sampling
    x0 = torch.floor(x).to(torch.int64)
    x1 = x0 + 1

    # Clip values
    x0 = torch.clamp(x0, 0, width - 1)
    x1 = torch.clamp(x1, 0, width - 1)

    # Batch effect
    # r = torch.arange(n_batch).repeat(outsize)
    # r = torch.arange(n_batch).repeat(outsize, 1).t().flatten()
    r = torch.arange(n_batch).repeat_interleave(outsize)

    # Index
    y0 = data[r, x0, :]
    y1 = data[r, x1, :]

    # Interpolation weights
    xd = (x - x0.to(torch.float32)).reshape((-1, 1))

    # Do interpolation
    y = y0 * (1 - xd) + y1 * xd

    newdata = torch.reshape(y, (n_batch, outsize, n_channels))

    return newdata


def interpolate_grid(data):
    # Problem size
    n_batch, width = data.shape

    # Rescale to interval [0-1] per batch
    # xmin, xmax = data[:,0].view(-1,1), data[:,-1].view(-1,1)
    # xdelta = xmax - xmin
    # data = (data - xmin) / xdelta

    # Extract points
    x = data.flatten()

    # Scale to domain
    x = x * (width - 1)

    # Do sampling
    x0 = torch.floor(x).to(torch.int64)
    x1 = x0 + 1

    # Clip values
    x0 = torch.clamp(x0, 0, width - 1)
    x1 = torch.clamp(x1, 0, width - 1)

    # Batch effect
    # r = torch.arange(n_batch).repeat_interleave(width)
    r = torch.arange(n_batch).view(-1, 1).repeat(1, width).flatten()

    # Index
    y0 = data[r, x0]
    y1 = data[r, x1]

    # Interpolation weights
    xd = (x - x0).to(torch.float32)

    # Do interpolation
    y = y0 * (1 - xd) + y1 * xd

    newdata = torch.reshape(y, (n_batch, width))

    # newdata = xmin + newdata * xdelta
    return newdata



# %% SETUP

import torch

# %% COMPARE EQUAL TO ZERO
eps = torch.finfo(torch.float32).eps


def cmpf(x, y):
    return torch.abs(x - y) < eps


def cmpf0(x):
    return torch.abs(x) < eps


# %% BATCH EFFECT


def batch_effect(x, theta):
    if x.ndim == 1:
        n_batch = theta.shape[0]
        n_points = x.shape[-1]
        x = torch.broadcast_to(x, (n_batch, n_points))  # .flatten()
    return x.flatten()


# %% FUNCTIONS


def get_affine(x, theta, params):
    if params.precomputed:
        return params.A, params.r
    else:
        n_batch = theta.shape[0]
        n_points = x.shape[-1]
        repeat = int(n_points / n_batch)

        # r = np.broadcast_to(np.arange(n_batch), [n_points, n_batch]).T
        # NOTE: here we suppose batch effect has been already executed
        r = torch.arange(n_batch).repeat_interleave(repeat).long().to(x.device)

        A = params.B.mm(theta.T).T.reshape(n_batch, -1, 2).to(x.device)
        return A, r


def precompute_affine(x, theta, params):
    params = params.copy()
    params.precomputed = False
    params.A, params.r = get_affine(x, theta, params)
    params.precomputed = True
    return params


def right_boundary(c, params):
    xmin, xmax, nc = params.xmin, params.xmax, params.nc
    return xmin + (c + 1) * (xmax - xmin) / nc + eps


def left_boundary(c, params):
    xmin, xmax, nc = params.xmin, params.xmax, params.nc
    return xmin + c * (xmax - xmin) / nc - eps


def get_cell(x, params):
    xmin, xmax, nc = params.xmin, params.xmax, params.nc

    c = torch.floor((x - xmin) / (xmax - xmin) * nc)
    c = torch.clamp(c, 0, nc - 1).long()
    return c


def get_velocity(x, theta, params):
    A, r = get_affine(x, theta, params)
    c = get_cell(x, params)
    a = A[r, c, 0]
    b = A[r, c, 1]
    return a * x + b


def calc_velocity(grid, theta, params):
    grid = batch_effect(grid, theta)
    v = get_velocity(grid, theta, params)
    return v.reshape(theta.shape[0], -1)


def get_psi(x, t, theta, params):
    A, r = get_affine(x, theta, params)
    c = get_cell(x, params)
    a = A[r, c, 0]
    b = A[r, c, 1]

    cond = cmpf0(a)
    x1 = x + t * b
    eta = torch.exp(t * a)
    x2 = eta * x + (b / a) * (eta - 1.0)
    # x2 = torch.exp(t * a) * (x + (b / a)) - (b / a)
    psi = torch.where(cond, x1, x2)
    return psi


def get_hit_time(x, theta, params):

    thit = torch.empty_like(x)
    valid = torch.full_like(x, True, dtype=bool)

    c = get_cell(x, params)
    A, r = get_affine(x, theta, params)

    a = A[r, c, 0]
    b = A[r, c, 1]

    v = a * x + b
    cond1 = cmpf0(v)

    cc = c + torch.sign(v)
    cond2 = torch.logical_or(cc < 0, cc >= params.nc)
    xc = torch.where(v > 0, right_boundary(c, params), left_boundary(c, params))

    vc = a * xc + b
    cond3 = cmpf0(vc)
    cond4 = torch.sign(v) != torch.sign(vc)
    cond5 = torch.logical_or(xc == params.xmin, xc == params.xmax)

    cond = cond1 | cond2 | cond3 | cond4 | cond5
    thit[~cond] = torch.where(
        cmpf0(a[~cond]), (xc[~cond] - x[~cond]) / b[~cond], torch.log(vc[~cond] / v[~cond]) / a[~cond],
    )
    thit[cond] = float("inf")
    return thit, xc, cc


def get_phi_numeric(x, t, theta, params):
    nSteps2 = params.nSteps2

    yn = x
    deltaT = t / nSteps2
    for j in range(nSteps2):
        midpoint = yn + deltaT / 2 * get_velocity(yn, theta, params)
        yn = yn + deltaT * get_velocity(midpoint, theta, params)
    return yn


# %% INTEGRATION


def integrate_numeric(x, theta, params, time=1.0):
    # setup
    x = batch_effect(x, theta)
    n_batch = theta.shape[0]
    t = time
    params = precompute_affine(x, theta, params)

    # computation
    xPrev = x
    nSteps1 = params.nSteps1
    deltaT = t / nSteps1
    for j in range(nSteps1):
        c = get_cell(xPrev, params)
        xTemp = get_psi(xPrev, deltaT, theta, params)
        cTemp = get_cell(xTemp, params)

        xNum = get_phi_numeric(xPrev, deltaT, theta, params)
        xPrev = torch.where(c == cTemp, xTemp, xNum)
    return xPrev.reshape((n_batch, -1))


def integrate_closed_form(x, theta, params, time=1.0):
    # setup
    x = batch_effect(x, theta)
    t = torch.ones_like(x) * time
    params = precompute_affine(x, theta, params)
    n_batch = theta.shape[0]

    # computation
    phi = torch.empty_like(x)
    done = torch.full_like(x, False, dtype=bool)

    c = get_cell(x, params)
    cont = 0
    while True:
        thit, xc, cc = get_hit_time(x, theta, params)
        psi = get_psi(x, t, theta, params)

        valid = thit > t
        phi[~done] = psi
        done[~done] = valid

        if torch.all(valid):
            return phi.reshape((n_batch, -1))

        params.r = params.r[~valid]
        x = xc[~valid]
        c = cc[~valid]
        t = (t - thit)[~valid]

        cont += 1
        if cont > params.nc:
            raise BaseException
    return None


# %% DERIVATIVE


def derivative_numeric(x, theta, params, time=1.0, h=1e-3):
    # setup
    n_points = x.shape[-1]
    n_batch = theta.shape[0]
    d = theta.shape[1]

    # computation
    der = torch.empty((n_batch, n_points, d), device=x.device)

    phi_1 = integrate_numeric(x, theta, params, time)
    for k in range(d):
        theta2 = theta.clone()
        theta2[:, k] += h
        phi_2 = integrate_numeric(x, theta2, params, time)

        der[:, :, k] = (phi_2 - phi_1) / h

    return phi_1, der


def derivative_closed_form(x, theta, params, time=1.0):
    # setup
    n_points = x.shape[-1]
    n_batch = theta.shape[0]
    d = theta.shape[1]

    # computation
    result = integrate_closed_form_trace(x, theta, params, time)
    phi = result[:, 0].reshape((n_batch, -1))
    tm = result[:, 1]
    cm = result[:, 2]

    # setup
    x = batch_effect(x, theta)
    params = precompute_affine(x, theta, params)

    der = torch.empty((n_batch, n_points, d), device=x.device)
    for k in range(d):
        dthit_dtheta_cum = torch.zeros_like(x)

        xm = x.clone()
        c = get_cell(x, params)
        while True:
            valid = c == cm
            if torch.all(valid):
                break
            step = torch.sign(cm - c)
            dthit_dtheta_cum[~valid] -= derivative_thit_theta(xm, theta, k, params)[~valid]
            xm[~valid] = torch.where(step == 1, right_boundary(c, params), left_boundary(c, params))[~valid]
            c = c + step

        dpsi_dtheta = derivative_psi_theta(xm, tm, theta, k, params)
        dpsi_dtime = derivative_phi_time(xm, tm, theta, k, params)
        dphi_dtheta = dpsi_dtheta + dpsi_dtime * dthit_dtheta_cum
        der[:, :, k] = dphi_dtheta.reshape(n_batch, n_points)

    return phi, der


def derivative_psi_theta(x, t, theta, k, params):
    A, r = get_affine(x, theta, params)
    A = A.double()  # NOTE: double precision is necessary

    c = get_cell(x, params)
    a = A[r, c, 0]
    b = A[r, c, 1]

    ak = params.B[2 * c, k]
    bk = params.B[2 * c + 1, k]

    cond = cmpf0(a)
    d1 = t * (x * ak + bk)
    d2 = ak * t * torch.exp(t * a) * (x + b / a) + (torch.exp(t * a) - 1) * (bk * a - ak * b) / a ** 2
    d1 = d1.double()
    d2 = d2.double()
    dpsi_dtheta = torch.where(cond, d1, d2)
    return dpsi_dtheta


def derivative_phi_time(x, t, theta, k, params):
    A, r = get_affine(x, theta, params)
    c = get_cell(x, params)
    a = A[r, c, 0]
    b = A[r, c, 1]

    cond = cmpf0(a)
    d1 = b
    d2 = torch.exp(t * a) * (a * x + b)
    dpsi_dtime = torch.where(cond, d1, d2)
    return dpsi_dtime


def derivative_thit_theta(x, theta, k, params):
    A, r = get_affine(x, theta, params)
    A = A.double()  # NOTE: double precision is necessary

    c = get_cell(x, params)
    a = A[r, c, 0]
    b = A[r, c, 1]

    ak = params.B[2 * c, k]
    bk = params.B[2 * c + 1, k]

    v = get_velocity(x, theta, params)

    xc = torch.where(v >= 0, right_boundary(c, params), left_boundary(c, params))

    cond = cmpf0(a)
    d1 = (x - xc) * bk / b ** 2
    d2 = -ak * torch.log((a * xc + b) / (a * x + b)) / a ** 2
    d3 = (x - xc) * (bk * a - ak * b) / (a * (a * x + b) * (a * xc + b))
    dthit_dtheta = torch.where(cond, d1, d2 + d3)
    return dthit_dtheta


# %% TRANSFORMATION


def integrate_closed_form_trace(x, theta, params, time=1.0):
    # setup
    x = batch_effect(x, theta)
    t = torch.ones_like(x) * time
    params = precompute_affine(x, theta, params)
    n_batch = theta.shape[0]

    # computation
    result = torch.empty((*x.shape, 3), device=x.device)
    done = torch.full_like(x, False, dtype=bool)

    c = get_cell(x, params)
    cont = 0
    while True:
        thit, xc, cc = get_hit_time(x, theta, params)
        psi = get_psi(x, t, theta, params)

        valid = thit > t
        result[~done] = torch.stack((psi, t, c)).T
        done[~done] = valid

        if torch.all(valid):
            return result

        params.r = params.r[~valid]
        x = xc[~valid]
        c = cc[~valid]
        t = (t - thit)[~valid]

        cont += 1
        if cont > params.nc:
            raise BaseException
    return None


def derivative_closed_form_trace(result, x, theta, params):
    # setup
    n_points = x.shape[-1]
    n_batch = theta.shape[0]
    d = theta.shape[1]

    # computation
    # result = integrate_closed_form_trace(x, theta, params, time)
    phi = result[:, 0].reshape((n_batch, -1))
    tm = result[:, 1]
    cm = result[:, 2]

    # setup
    x = batch_effect(x, theta)
    params = precompute_affine(x, theta, params)

    der = torch.empty((n_batch, n_points, d), device=x.device)
    for k in range(d):
        dthit_dtheta_cum = torch.zeros_like(x)

        xm = x.clone()
        c = get_cell(x, params)
        while True:
            valid = c == cm
            if torch.all(valid):
                break
            step = torch.sign(cm - c)
            dthit_dtheta_cum[~valid] -= derivative_thit_theta(xm, theta, k, params)[~valid]
            xm[~valid] = torch.where(step == 1, right_boundary(c, params), left_boundary(c, params))[~valid]
            c = c + step

        dpsi_dtheta = derivative_psi_theta(xm, tm, theta, k, params)
        dpsi_dtime = derivative_phi_time(xm, tm, theta, k, params)
        dphi_dtheta = dpsi_dtheta + dpsi_dtime * dthit_dtheta_cum
        der[:, :, k] = dphi_dtheta.reshape(n_batch, n_points)

    return der


def derivative_numeric_trace(phi_1, x, theta, params, time=1.0, h=1e-3):
    # setup
    n_points = x.shape[-1]
    n_batch = theta.shape[0]
    d = theta.shape[1]

    # computation
    der = torch.empty((n_batch, n_points, d), device=x.device)

    # phi_1 = integrate_numeric(x, theta, params)
    for k in range(d):
        theta[:, k] += h
        phi_2 = integrate_numeric(x, theta, params, time)
        theta[:, k] -= h

        der[:, :, k] = (phi_2 - phi_1) / h
    return der


# %% GRADIENT SPACE

def derivative_space_numeric(x, theta, params, time=1.0, h=1e-3):
    # setup
    n_points = x.shape[-1]
    n_batch = theta.shape[0]
    d = theta.shape[1]

    # phi_1 = integrate_numeric(x, theta, params, time)
    # phi_2 = integrate_numeric(x+h, theta, params, time)
    # der = (phi_2 - phi_1) / h
    # return phi_1, der

    # computation
    xe = torch.cat([x, x+h])

    phi = integrate_numeric(xe, theta, params, time)
    phi_1, phi_2 = torch.split(phi, n_points, dim=1)
    der = (phi_2 - phi_1) / h

    return phi_1, der

def derivative_space_closed_form(x, theta, params, time=1.0):
    # setup
    n_points = x.shape[-1]
    n_batch = theta.shape[0]
    d = theta.shape[1]

    # computation
    t = torch.ones_like(x) * time
    result = integrate_closed_form_trace(x, theta, params, time)
    phi = result[:, 0].reshape((n_batch, -1))
    tm = result[:, 1]
    cm = result[:, 2]

    # setup
    x = batch_effect(x, theta)
    params = precompute_affine(x, theta, params)

    dthit_dx = torch.zeros_like(x)
    dpsi_dx = torch.zeros_like(x)

    c = get_cell(x, params)
    valid = c == cm
    # dpsi_dx only on first valid cell
    dpsi_dx[valid] = derivative_psi_x(x, t, theta, params)[valid]
    # dthit_dx only on first non valid cell
    dthit_dx[~valid] = derivative_thit_x(x, t, theta, params)[~valid]

    xm = x.clone()
    while True:
        valid = c == cm
        if torch.all(valid):
            break
        step = torch.sign(cm - c)
        xm[~valid] = torch.where(step == 1, right_boundary(c, params), left_boundary(c, params))[~valid]
        c = c + step

    dpsi_dtime = derivative_psi_t(xm, tm, theta, params)
    dphi_dx = dpsi_dx + dpsi_dtime * dthit_dx
    dphi_dx = dphi_dx.reshape(n_batch, n_points)

    return phi, dphi_dx


def derivative_thit_x(x, t, theta, params):
    A, r = get_affine(x, theta, params)
    c = get_cell(x, params)
    a = A[r, c, 0]
    b = A[r, c, 1]

    return 1.0 / (a*x + b)

def derivative_psi_x(x, t, theta, params):
    A, r = get_affine(x, theta, params)
    c = get_cell(x, params)
    a = A[r, c, 0]
    b = A[r, c, 1]

    return torch.exp(t * a)

def derivative_psi_t(x, t, theta, params):
    A, r = get_affine(x, theta, params)
    c = get_cell(x, params)
    a = A[r, c, 0]
    b = A[r, c, 1]
    return torch.exp(t * a) * (a * x + b)


import torch

from .interpolation import interpolate
from .transformer import transformer, gradient, gradient_space
from .transformer import get_cell, calc_velocity
from .transformer import interpolate_grid

# %%
def assert_version():
    numbers = torch.__version__.split(".")
    version = float(numbers[0] + "." + numbers[1])
    assert (
        version >= 1.0
    ), """ You are using a older installation of pytorch, please install 1.0.0
            or newer """


# %%
half = torch.float16
single = torch.float32
double = torch.float64


def to(x, dtype=torch.float32, device=None):
    if type(device) == str:
        device = torch.device("cuda") if device == "gpu" else torch.device("cpu")
    if torch.is_tensor(x):
        return x.detach().clone().type(dtype).to(device)
    return torch.tensor(x, dtype=dtype, device=device)


# %%
def tonumpy(x):
    return x.cpu().numpy()


# %%
def check_device(x, device_name):
    return (x.is_cuda) == (device_name == "gpu")


# %%
def backend_type():
    return torch.Tensor


# %%
def sample_transformation(d, n_sample=1, mean=None, cov=None, device="cpu"):
    device = torch.device("cpu") if device == "cpu" else torch.device("cuda")
    mean = torch.zeros(d, dtype=torch.float32, device=device) if mean is None else mean
    cov = torch.eye(d, dtype=torch.float32, device=device) if cov is None else cov
    distribution = torch.distributions.MultivariateNormal(mean, cov)
    return distribution.sample((n_sample,)).to(device)


# %%
def identity(d, n_sample=1, epsilon=0, device="cpu"):
    assert epsilon >= 0, "epsilon need to be larger than 0"
    device = torch.device("cpu") if device == "cpu" else torch.device("cuda")
    return torch.zeros(n_sample, d, dtype=torch.float32, device=device) + epsilon


# %%
def uniform_meshgrid(xmin, xmax, n_points, device="cpu"):
    device = torch.device("cpu") if device == "cpu" else torch.device("cuda")
    return torch.linspace(xmin, xmax, n_points, dtype=torch.float32, device=device)


# %%
def exp(*args, **kwargs):
    return torch.exp(*args, **kwargs)


def linspace(*args, **kwargs):
    return torch.linspace(*args, **kwargs)


def meshgrid(*args, **kwargs):
    return torch.meshgrid(*args, **kwargs)


def matmul(*args, **kwargs):
    return torch.matmul(*args, **kwargs)


def max(*args, **kwargs):
    return torch.max(*args, **kwargs)


def ones(*args, **kwargs):
    return torch.ones(*args, **kwargs)


def pdist(c):
    x, y = torch.meshgrid(c, c)
    return torch.abs(x - y)



import numpy as np


def interpolate(data, grid, outsize):
    # Problem size
    n_batch, width, n_channels = data.shape

    # Extract points
    x = grid.flatten()

    # Scale to domain
    x = x * (width - 1)

    # Do sampling
    x0 = np.floor(x).astype(np.int32)
    x1 = x0 + 1

    # Clip values
    x0 = np.clip(x0, 0, width - 1)
    x1 = np.clip(x1, 0, width - 1)

    # Batch effect
    r = np.arange(n_batch).repeat(outsize)

    # Index
    y0 = data[r, x0, :]
    y1 = data[r, x1, :]

    # Interpolation weights
    xd = (x - x0.astype(np.float32)).reshape((-1, 1))

    # Do interpolation
    y = y0 * (1 - xd) + y1 * xd

    newdata = np.reshape(y, (n_batch, outsize, n_channels))
    return newdata


def interpolate_grid(data):
    # Problem size
    n_batch, width = data.shape

    # Extract points
    x = data.flatten()

    # Scale to domain
    x = x * (width - 1)

    # Do sampling
    x0 = np.floor(x).astype(np.int32)
    x1 = x0 + 1

    # Clip values
    x0 = np.clip(x0, 0, width - 1)
    x1 = np.clip(x1, 0, width - 1)

    # Batch effect
    r = np.arange(n_batch).repeat(width)

    # Index
    y0 = data[r, x0]
    y1 = data[r, x1]

    # Interpolation weights
    xd = x - x0.astype(np.float32)

    # Do interpolation
    y = y0 * (1 - xd) + y1 * xd

    newdata = np.reshape(y, (n_batch, width))
    return newdata



import numpy as np
import matplotlib.pyplot as plt
from scipy.linalg import null_space


class Tessellation:
    """Regular grid tessellation"""
    
    def __init__(self, nc, xmin=0, xmax=1, zero_boundary=True, basis="rref"):
        self.nc = nc
        self.nv = nc + 1
        self.ns = nc - 1

        self.xmin = xmin
        self.xmax = xmax
        self.xr = self.xmax - self.xmin

        self.zero_boundary = zero_boundary
        self.ord = np.inf
        # self.ord = 2

        if basis == "svd":
            self.L = self.constrain_matrix()
            self.B = self.basis_svd()
        elif basis == "rref":
            self.B = self.generate_basis_rref()
        elif basis == "sparse":
            self.B = self.generate_basis_sparse()
        elif basis == "qr":
            self.L = self.constrain_matrix()
            self.B = self.basis_qr()

        self.D, self.d = self.B.shape

    def cell_centers(self):
        h = self.xr / self.nc
        return np.linspace(self.xmin, self.xmax - h, self.nc) + h / 2

    def constrain_matrix(self):
        vertices = np.linspace(self.xmin, self.xmax, self.nv)
        shared_vertices = vertices[1:-1]

        rows = self.ns
        cols = 2 * self.nc
        L = np.zeros((rows, cols))
        for i, x in enumerate(shared_vertices):
            L[i, 2 * i : 2 * (i + 2)] = [x, 1, -x, -1]

        if self.zero_boundary:
            Ltemp = self.zero_boundary_constrains()
            L = np.vstack((L, Ltemp))

        return L

    def zero_boundary_constrains(self):
        L = np.zeros((2, 2 * self.nc))
        L[0, :2] = [-self.xmin, -1]
        L[1, -2:] = [-self.xmax, -1]
        return L

    def generate_basis_rref(self):
        if self.zero_boundary:
            return self.basis_rref_zb()
        else:
            return self.basis_rref()

    def basis_svd(self):
        B = null_space(self.L)
        # B = B / np.linalg.norm(B, ord=self.ord, axis=0)
        return B

    def basis_qr(self):
        B = self.qr_null(self.L)
        # B = B / np.linalg.norm(B, ord=self.ord, axis=0)
        return B

    def qr_null(self, A, tol=None):
        from scipy.linalg import qr

        Q, R, P = qr(A.T, mode="full", pivoting=True)
        tol = np.max(A) * np.finfo(R.dtype).eps if tol is None else tol
        rnk = min(A.shape) - np.abs(np.diag(R))[::-1].searchsorted(tol)
        return Q[:, rnk:].conj()

    # with zero boundary
    def basis_rref_zb(self):
        rows = self.nc - 1
        cols = 2 * self.nc
        B = np.zeros((rows, cols))

        a = self.xmin
        b = self.xmax
        n = self.nc
        s = (b - a) / n
        B[:, 0] = a + s
        B[:, 1] = -a * (a + s)
        for k in range(1, self.nc):
            B[k - 1, 2 : 2 * k : 2] = s
            B[k - 1, 2 * k] = -(a + k * s)
            B[k - 1, 2 * k + 1] = (a + k * s) * (a + (k + 1) * s)

        # normalize
        B = B.T / np.linalg.norm(B, ord=self.ord, axis=1)
        return B

    # without zero boundary
    def basis_rref_backup(self):
        rows = self.nc + 1
        cols = 2 * self.nc
        B = np.zeros((rows, cols))

        a = self.xmin
        b = self.xmax
        n = self.nc
        s = (b - a) / n

        B[0, 0] = 1
        B[0, 1] = -(a + s)
        B[-2, ::2] = 1
        B[-1, :-2:2] = 1
        B[-1, -1] = a + (n - 1) * s
        for k in range(1, n - 1):
            B[k, : 2 * k : 2] = s
            B[k, 2 * k] = -(a + k * s)
            B[k, 2 * k + 1] = (a + k * s) * (a + (k + 1) * s)

        # normalize
        B = B.T / np.linalg.norm(B, axis=1)
        return B

    def basis_rref(self):
        rows = self.nc + 1
        cols = 2 * self.nc
        B = np.zeros((rows, cols))

        a = self.xmin
        b = self.xmax
        n = self.nc
        s = (b - a) / n

        B[0, 0] = -1
        B[0, 1] = a + s
        B[-2, ::2] = 1
        # B[-1, :-2:2] = 1
        B[-1, -2] = 1
        B[-1, -1] = -(a + (n - 1) * s)
        for k in range(1, n - 1):
            B[k, : 2 * k : 2] = s
            B[k, 2 * k] = -(a + k * s)
            B[k, 2 * k + 1] = (a + k * s) * (a + (k + 1) * s)

        # return B.T

        # normalize
        B = B.T / np.linalg.norm(B, ord=self.ord, axis=1)
        if self.ord == np.inf:
            B[::2, -2] = 1 / n
        return B

    # with zero boundary
    def basis_rref_zb_new(self):
        rows = self.nc - 1
        cols = 2 * self.nc
        B = np.zeros((rows, cols))

        a = self.xmin
        b = self.xmax
        n = self.nc
        B[:, 0] = (n - 1) * a + b
        B[:, 1] = -a * ((n - 1) * a + b)
        for k in range(1, self.nc):
            B[k - 1, 2 : 2 * k : 2] = b - a
            B[k - 1, 2 * k] = -((n - k) * a + k * b)
            B[k - 1, 2 * k + 1] = ((n - k) * a + k * b) * ((n - k - 1) * a + (k + 1) * b) / n

        # normalize
        B = B.T / np.linalg.norm(B, axis=1)
        return B

    # without zero boundary
    def basis_rref_new(self):
        rows = self.nc + 1
        cols = 2 * self.nc
        B = np.zeros((rows, cols))

        a = self.xmin
        b = self.xmax
        n = self.nc

        B[0, 0] = n
        B[0, 1] = -((n - 1) * a + b)
        B[-2, ::2] = n
        B[-1, :-2:2] = n
        B[-1, -1] = a + (n - 1) * b
        for k in range(1, n - 1):
            B[k, : 2 * k : 2] = b - a
            B[k, 2 * k] = -((n - k) * a + k * b)
            B[k, 2 * k + 1] = ((n - k) * a + k * b) * ((n - k - 1) * a + (k + 1) * b) / n

        # normalize
        B = B.T / np.linalg.norm(B, axis=1)
        return B

    def generate_basis_sparse(self):
        if self.zero_boundary:
            return self.basis_sparse_zb()
        else:
            return self.basis_sparse()

    def basis_sparse(self):
        rows = 2 * self.nc
        cols = self.nv
        B = np.zeros((rows, cols))

        s = (self.xmax - self.xmin) / self.nc

        r = np.arange(0, rows, 2)
        c = np.arange(cols - 1)

        B[r, c] = -1
        B[r, c + 1] = 1
        B[r + 1, c] = np.arange(self.xmin + s, self.xmax + s, s)
        B[r + 1, c + 1] = -np.arange(self.xmin, self.xmax, s)

        B = B / s
        B = B / np.linalg.norm(B, ord=self.ord, axis=0)

        return B

    def basis_sparse_zb(self):
        B = self.basis_sparse()[:, 1:-1]
        B = B / np.linalg.norm(B, ord=self.ord, axis=0)
        return B

    def plot_basis(self):
        plt.figure()
        plt.spy(self.B)



import os

# %%


class Parameters:
    def __repr__(self):
        return str(self.__dict__)

    def __copy__(self):
        newone = type(self)()
        newone.__dict__.update(self.__dict__)
        return newone

    def copy(self):
        newone = type(self)()
        newone.__dict__.update(self.__dict__)
        return newone


# %%
def get_dir(file):
    """ Get directory of the input file """
    return os.path.dirname(os.path.realpath(file))


# %%


class methods:
    closed_form = "closed_form"
    numeric = "numeric"

    @staticmethod
    def check(method):
        if method is not None:
            assert method in [methods.closed_form, methods.numeric], (
                "Unknown method, choose between " + methods.closed_form + " or " + methods.numeric
            )

    @staticmethod
    def default(method=None):
        if method is not None:
            return method
        else:
            return methods.closed_form
